#cs ----------------------------------------------------------------------------

 AutoIt Version: 3.3.6.1
 Author:         Matwachich

 Script Function:
	Script d'exemple d'utilisation de threads
	- Un clique sur New lance un nouveau Thread, celui-ci apprait alors dans la
		liste
	- Le thread commence  compter les secondes, quand on clique sur le thread
		dans la liste, ses informations s'affichent
	- On peut alors le commander: Pause, Start, Et Quit

#ce ----------------------------------------------------------------------------
#NoTrayIcon

#include <ButtonConstants.au3>
#include <EditConstants.au3>
#include <GUIConstantsEx.au3>
#include <GUIListBox.au3>
#include <StaticConstants.au3>
#include <WindowsConstants.au3>
#Include <GuiListBox.au3>
#Include <Array.au3>

#include "Thread.au3"

; PS: un ThreadID est tout simplement sont index dans ce Array (le x)
Global $__Threads[1][2] = [[0, ""]] ; array qui va contenir tous les threads: [x][0] = PID, [x][1] = Count

Global $var, $CurrDisplayedID = 0

#Region ### START Koda GUI section ### Form=
$GUI_Main = GUICreate("Thread Example", 284, 260, -1, -1)
GUICtrlCreateLabel("Threads List", 6, 6, 62, 17)
; ---
$List = GUICtrlCreateList("", 6, 24, 121, 197, BitOR($LBS_NOTIFY,$LBS_SORT,$LBS_NOINTEGRALHEIGHT,$WS_VSCROLL))
	$hList = GUICtrlGetHandle($List)
; ---
GUICtrlCreateGroup(" Thread Information ", 132, 18, 143, 135)
	GUICtrlCreateLabel("PID", 144, 42, 22, 17)
	$I_Pid = GUICtrlCreateInput("", 144, 60, 115, 21, BitOR($GUI_SS_DEFAULT_INPUT,$ES_READONLY,$ES_RIGHT))
	GUICtrlCreateLabel("Count", 144, 96, 32, 17)
	$I_Count = GUICtrlCreateInput("", 144, 114, 115, 21, BitOR($GUI_SS_DEFAULT_INPUT,$ES_READONLY,$ES_RIGHT))
GUICtrlCreateGroup("", -99, -99, 1, 1)
; ---
$B_New = GUICtrlCreateButton("New", 30, 228, 75, 25)
$B_Start = GUICtrlCreateButton("Start", 168, 162, 75, 25)
$B_Pause = GUICtrlCreateButton("Pause", 168, 192, 75, 25)
$B_Quit = GUICtrlCreateButton("Quit", 168, 222, 75, 25)
; ---
GUISetState(@SW_SHOW)
#EndRegion ### END Koda GUI section ###

While 1
	$nMsg = GUIGetMsg()
	Switch $nMsg
		Case $GUI_EVENT_CLOSE
			For $i = 1 To $__Threads[0][0]
				ProcessClose($__Threads[$i][0])
			Next
			Exit
		Case $B_New
			_Thread_Add()
		Case $B_Start
			$var = __ListBox_GetSelID()
			If $var Then _Thread_Resume($var)
		Case $B_Pause
			$var = __ListBox_GetSelID()
			If $var Then _Thread_Pause($var)
		Case $B_Quit
			_Thread_Del()
			; ---
		Case $List
			$var = __ListBox_GetSelID()
			If $var Then
				GuiCtrlSetData($I_Pid, $__Threads[$var][0])
				GUICtrlSetData($I_Count, $__Threads[$var][1])
				$CurrDisplayedID = $var
			EndIf
	EndSwitch
WEnd

; ##############################################################

; Lance un nouveau thread, le stock dans le Array, et ajoute une entre dans la liste
Func _Thread_Add()
	Local $pid = _CoProc('__Thread_Function()', '__Thread_RecvFunc')
	If Not $pid Then Return MsgBox(16, "Error", "Impossible de crer le thread")
	; ---
	Local $id = _Array_Add($pid)
	; ---
	_GUICtrlListBox_AddString($hList, "Thread N" & $id)
	Return 1
EndFunc

; Si un thread est slectionn dans la liste, alors on le ferme, on le supprime du Array et de la Liste
Func _Thread_Del()
	Local $sel = __ListBox_GetSelID()
	If Not $sel Then Return
	; ---
	ProcessClose($__Threads[$sel][0])
	_Array_Del($sel)
	; ---
	_GuiCtrlListBox_DeleteString($hList, "Thread N" & $sel)
	; ---
	If $sel = $CurrDisplayedID Then
		GuiCtrlSetData($I_Pid, "")
		GUICtrlSetData($I_Count, "")
	EndIf
	; ---
	Return 1
EndFunc

; ##############################################################

Func _Thread_Pause($id)
	Local $pid = $__Threads[$id][0]
	; ---
	_CoProc_Pause($pid)
EndFunc

Func _Thread_Resume($id)
	Local $pid = $__Threads[$id][0]
	; ---
	_CoProc_Resume($pid)
EndFunc

; ##############################################################

; Voici le thread!
Func __Thread_Function()
	Local $i = 0, $pause = 0
	While 1
		Sleep(1000)
		; ---
		$i += 1
		_CoProc_SendMsg($i)
	WEnd
EndFunc

; ##############################################################

; Voici la fonction que recevera les _CoProc_SendMsg des threads
Func __Thread_RecvFunc($iPid, $sData)
	For $i = 1 To $__Threads[0][0]
		If $iPid = $__Threads[$i][0] Then
			$__Threads[$i][1] = $sData
			; ---
			If $i = $CurrDisplayedID Then
				GUICtrlSetData($I_Count, $sData)
			EndIf
		EndIf
	Next
EndFunc

; ##############################################################
; Ces fonctions de gestion de array bizarement complxe sont pourtant essentielles
; pour que les ID ne se dcalent pas lors de la suppression d'un lment

Func _Array_Add($pid)
	_Array_Clean()
	; ---
	Local $id = 0
	; ---
	For $i = 1 To $__Threads[0][0]
		If $__Threads[$i][0] = 0 Then $id = $i
	Next
	; ---
	If Not $id Then
		Redim $__Threads[$__Threads[0][0] + 2][2]
		$__Threads[0][0] += 1
		$id = $__Threads[0][0]
	EndIf
	; ---
	$__Threads[$id][0] = $pid
	$__Threads[$id][1] = ""
	; ---
	Return $id
EndFunc

Func _Array_Del($id)
	If $id > $__Threads[0][0] Then Return
	; ---
	$__Threads[$id][0] = 0
	$__Threads[$id][1] = ""
	; ---
	_Array_Clean()
	Return 1
EndFunc

Func _Array_Clean()
	For $i = $__Threads[0][0] To 1 Step -1
		If $__Threads[$i][0] Then ExitLoop
		; ---
		_ArrayDelete($__Threads, $i)
		$__Threads[0][0] -= 1
	Next
EndFunc

; ##############################################################

; Retourne le ThreadID actuellement slectionn dans la Liste
Func __ListBox_GetSelID()
	Local $sel = _GUICtrlListBox_GetText($hList, _GUICtrlListBox_GetCurSel($hList))
	If Not $sel Then Return 0
	; ---
	Return StringTrimLeft($sel, StringLen("Thread N"))
EndFunc
